"""
Abstract websocket functionality.

"""

from __future__ import print_function
from __future__ import unicode_literals

from base64 import b64encode
from hashlib import sha1
import json
import logging
import os
import time

import six
from six import text_type
from six.moves.urllib.parse import urlparse

from . import constants
from . import errors
from . import events
from .frame import Frame
from .opcode import Opcode
from .response import Response
from .stream import WebsocketStream
from .session import WebsocketSession
from .status import Status


log = logging.getLogger('lomond')


class WebSocket(object):
    """IO independent websocket functionality.

    :param str url: A websocket URL, must have a ``ws://`` or ``wss://``
        protocol.
    :param dict proxies: A dict containing ``'http'`` or ``'https'``
        urls to a proxy server, or ``None`` to attempt to auto-detect
        proxies from environment. Pass an empty dict to disable proxy.
    :params list protocols: A list of supported protocols (defaults to
        no protocols).
    :params str agent: A user agent string to be sent in the header. The
        default uses the value ``USER_AGENT`` defined in
        :mod:`lomond.constants`.

    """

    class State(object):
        def __init__(self):
            self.stream = WebsocketStream()
            self.session = None
            self.key = b64encode(os.urandom(16))
            self.sent_request = False
            self.closing = False
            self.closed = False
            self.sent_close_time = None

    def __init__(self, url, proxies=None, protocols=None, agent=None):
        self.url = url
        self.proxies = self._detect_proxies() if proxies is None else proxies
        self.protocols = protocols or []
        self.agent = agent or constants.USER_AGENT
        self._headers = []
        _url = urlparse(url)
        self.scheme = _url.scheme
        self.host = _url.hostname
        self.port = (
            int(_url.port)
            if _url.port else
            (443 if self.scheme == 'wss' else 80)
        )
        self._host_port = "{}:{}".format(self.host, self.port)
        self.resource = _url.path or '/'
        if _url.query:
            self.resource = "{}?{}".format(self.resource, _url.query)

        self.state = self.State()

    @classmethod
    def _detect_proxies(cls):
        """Get proxy information form environment."""
        proxies = {
            'http': os.environ.get('HTTP_PROXY'),
            'https': os.environ.get('HTTPS_PROXY')
        }
        return proxies

    def __repr__(self):
        return "WebSocket('{}')".format(self.url)

    @property
    def is_secure(self):
        """Boolean that indicates if the websocket is over ssl (i.e. the
        `wss` protocol).

        """
        return self.scheme == 'wss'

    @property
    def is_closing(self):
        """Boolean that indicates if the websocket is in a closing
        state. No further messages may be sent when a websocket is
        closing.

        """
        return self.state.closing

    @property
    def is_active(self):
        """Boolean that indicates the socket is 'active' i.e. not in
        a closing state.

        """
        return not self.state.closing

    @property
    def sent_close_time(self):
        """The epoch time a close packet was sent (or None if no close
        packet has been sent).

        """
        return self.state.sent_close_time

    @property
    def is_closed(self):
        """Flag that indicates if the websocket is closed."""
        return self.state.closed

    @property
    def stream(self):
        return self.state.stream

    @property
    def session(self):
        return self.state.session

    @property
    def key(self):
        return self.state.key

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Close the session (and potentially a socket) on exit."""
        if self.session is not None:
            self.session.close()

    def add_header(self, header, value):
        """Add a custom header to the websocket request.

        :param bytes header: Name of the header.
        :param bytes value: Value of the header.

        """
        if not isinstance(header, bytes):
            raise TypeError("'header' must be bytes")
        if not isinstance(value, bytes):
            raise TypeError("'value' must be bytes")
        self._headers.append((header, value))

    def connect(self,
                session_class=WebsocketSession,
                poll=5.0,
                ping_rate=30.0,
                ping_timeout=None,
                auto_pong=True,
                close_timeout=30.0):
        """Connect the websocket to a session.

        :param session_class: An object to manage the *session*. This
            object is an extension mechanism that will allow the
            WebSocket to be *driven* by different back-ends. For now,
            treat it as an implementation detail and leave it as the
            default.
        :param float poll: Rate (in seconds) that poll events should be
            generated.
        :param float ping_rate: Rate that ping packets should be sent.
            Set to `0` to disable auto pings.
        :param float ping_timeout: Maximum time in seconds to wait for a
            pong response before disconnecting. Set to `None` (default)
            to disable. If set, double `ping_rate` would be a good
            starting point.
        :param bool auto_pong: Enable (default) automatic response to
            ping events.
        :param float close_timeout: Seconds to wait for server to
            respond to a close packet, before closing the socket. Set to
            `None` or `0` to disable the timeout.
        :returns: An iterable of :class:`~lomond.event.Event` instances.

        """
        self.reset()
        self.state.session = session = session_class(self)
        run_generator = session.run(
            poll=poll,
            ping_rate=ping_rate,
            ping_timeout=ping_timeout,
            auto_pong=auto_pong,
            close_timeout=close_timeout
        )
        return run_generator


    def reset(self):
        """Reset the state."""
        self.state = self.State()

    __iter__ = connect

    def close(self, code=None, reason=None):
        """Close the websocket.

        :param int code: A closing code, which should probably be one of
            the enumerations in :class:`lomond.status.Status` or a valid
            value as specified in
            https://tools.ietf.org/html/rfc6455#section-7.4
        :param str reason: A short descriptive reason why the websocket
            is closing. This value is intended for the remote end to
            help in debugging.

        .. note::
            Closing the websocket won't exit the main loop immediately;
            it will put the websocket in to a *closing* state while it
            waits for the server to echo back a close packet. No data
            may be sent by the application when the websocket is
            closing.

        """
        if self.is_closed:
            log.debug('%r already closed', self)
        else:
            if code is None:
                code = Status.NORMAL
            if reason is None:
                reason = b'goodbye'
            self._send_close(code, reason)
            self.state.closing = True
            self.state.sent_close_time = time.time()

    def _on_close(self, message):
        """Close logic generator."""
        if message.code in Status.invalid_codes:
            raise errors.ProtocolError(
                'reserved close code ({})',
                message.code
            )
        if self.is_closed:
            return
        if self.is_closing:
            yield events.Closed(message.code, message.reason)
            self.state.closing = False
            self.state.closed = True
        else:
            yield events.Closing(message.code, message.reason)
            self.close(message.code, message.reason)
            self.state.closing = True

    def on_disconnect(self):
        """Called on disconnect."""
        self.state.session.close()
        self.state.closing = False
        self.state.closed = True

    def feed(self, data):
        """Feed with data from the socket, and yield any events.

        This method is called by the Session object, and is not needed
        for normal use.

        :param bytes data: data received over a socket.

        """
        if self.is_closed:
            return
        try:
            for message in self.stream.feed(data):
                if isinstance(message, Response):
                    response = message
                    try:
                        protocol, extensions = self.on_response(response)
                    except errors.HandshakeError as error:
                        self.on_disconnect()
                        yield events.Rejected(response, six.text_type(error))
                        break
                    else:
                        yield events.Ready(response, protocol, extensions)
                else:
                    if message.is_close:
                        for event in self._on_close(message):
                            yield event
                    elif message.is_ping:
                        yield events.Ping(message.data)
                    elif message.is_pong:
                        yield events.Pong(message.data)
                    elif message.is_binary:
                        yield events.Binary(message.data)
                    elif message.is_text:
                        yield events.Text(message.text)
                if self.is_closed:
                    break

        except errors.CriticalProtocolError as error:
            # An error that warrants an immediate disconnect.
            # Usually invalid unicode.
            log.debug('critical protocol error; %s', error)
            self.on_disconnect()

        except errors.ProtocolError as error:
            # A violation of the protocol that allows for a graceful
            # disconnect.
            log.debug('protocol error; %s', error)
            self.close(Status.PROTOCOL_ERROR, six.text_type(error))
            self.on_disconnect()

        except GeneratorExit:
            # The generator has exited prematurely, due to an exception
            # handling the event.
            log.warning('disconnecting websocket')
            self.on_disconnect()

    def build_request(self):
        """Get the websocket request (in bytes).

        This method is called from the session, and should not be
        invoked explicitly.

        """
        request = [
            "GET {} HTTP/1.1".format(self.resource).encode('utf-8')
        ]
        version = '{}'.format(constants.WS_VERSION)
        headers = self._headers[:]

        headers.extend([
            (b'Host', self._host_port.encode('utf-8')),
            (b'Upgrade', b'websocket'),
            (b'Connection', b'Upgrade'),
            (b'Sec-WebSocket-Key', self.key),
            (b'Sec-WebSocket-Version', version.encode('utf-8')),
            (b'User-Agent', self.agent.encode('utf-8')),
        ])
        if self.protocols:
            protocols = ", ".join(self.protocols).encode('utf-8')
            headers.append((b'Sec-WebSocket-Protocol', protocols))
        for header, value in headers:
            request.append(header + b': ' + value)
        request.append(b'\r\n')
        request_bytes = b'\r\n'.join(request)
        return request_bytes

    def on_response(self, response):
        """Called when the HTTP response has been received."""

        if response.status_code != 101:
            raise errors.HandshakeError(
                'Websocket upgrade failed (code={})',
                response.status_code
            )

        upgrade_header = response.get(b'upgrade', b'?').lower()
        if upgrade_header != b'websocket':
            raise errors.HandshakeError(
                "Can't upgrade to {}",
                upgrade_header.decode('utf-8', errors='replace')
            )

        accept_header = response.get(b'sec-websocket-accept', None)
        if accept_header is None:
            raise errors.HandshakeError(
                "No Sec-WebSocket-Accept header"
            )

        challenge = b64encode(
            sha1(self.key + constants.WS_KEY).digest()
        )

        if accept_header.lower() != challenge.lower():
            raise errors.HandshakeError(
                "Sec-WebSocket-Accept challenge failed"
            )

        protocol = response.get(b'sec-websocket-protocol')
        extensions = set(response.get_list(b'sec-websocket-extensions'))
        return protocol, extensions

    def send_ping(self, data=b''):
        """Send a ping packet.

        :param bytes data: Data to send in the ping message (must be <=
            125 bytes).
        :raises TypeError: If `data` is not bytes.
        :raises ValueError: If `data` is > 125 bytes.

        """
        if not isinstance(data, bytes):
            raise TypeError('data argument must be bytes')
        if len(data) > 125:
            raise ValueError('ping data should be <= 125 bytes')
        self.session.send(Opcode.PING, data)

    def send_pong(self, data):
        """Send a pong packet.

        :param bytes data: Data to send in the ping message (must be <=
            125 bytes).

        A *pong* may be sent in response to a ping, or unsolicited to
        keep the connection alive.

        :raises TypeError: If `data` is not bytes.
        :raises ValueError: If `data` is > 125 bytes.

        """
        if not isinstance(data, bytes):
            raise TypeError('data argument must be bytes')
        if len(data) > 125:
            raise ValueError('pong data should be <= 125 bytes')
        self.session.send(Opcode.PONG, data)

    def send_binary(self, data):
        """Send a binary message.

        :param bytes data: Binary data to send.
        :raises TypeError: If data is not bytes.

        """
        if not isinstance(data, bytes):
            raise TypeError('data argument must be bytes')
        self.session.send(Opcode.BINARY, data)

    def send_json(self, _obj=Ellipsis, **kwargs):
        """Encode an object as JSON and send a text message.

        The object to encode may be specified as a single positional
        argument OR if as keyword arguments which will be encoded as a
        JSON object. The following two lines will send the
        same JSON::

            websocket.send_json({'foo': 'bar'})
            websocket.send_json(foo='bar')

        :param obj: An object to be encoded as JSON.
        :raises TypeError: If `obj` could not be encoded as JSON.

        """
        if kwargs and _obj is not Ellipsis:
            raise ValueError(
                'send_json requires positional argument OR keyword arguments'
            )
        json_obj = json.dumps(_obj if _obj is not Ellipsis else kwargs)
        self.send_text(
            json_obj.decode('utf-8')
            if six.PY2
            else json_obj
        )

    def send_text(self, text):
        """Send a text message.

        :param str text: Text to send.
        :raises TypeError: If data is not str (or unicode on Py2).

        """
        if not isinstance(text, six.text_type):
            raise TypeError('text argument must not be bytes')
        self.session.send(Opcode.TEXT, text.encode('utf-8'))

    def _send_close(self, code, reason):
        """Send a close frame."""
        frame_bytes = Frame.build_close_payload(code, reason)
        try:
            self.session.send(Opcode.CLOSE, frame_bytes)
        except (errors.WebSocketUnavailable, errors.TransportFail):
            return False
        else:
            return True



"""
Functions related to masking Websocket frames.
https://tools.ietf.org/html/rfc6455#section-5.3

"""

import os
from functools import partial

import six


try:
    from wsaccel.xormask import XorMaskerSimple
except ImportError:
    XorMaskerSimple = None


make_masking_key = partial(os.urandom, 4)


if XorMaskerSimple is not None:
    # Fast C version (works with Py2 and Py3)
    def mask(masking_key, data):
        return XorMaskerSimple(masking_key).process(data)

else:
    # This is about 60 (!) times faster than a simple loop
    # -------------------------------------------------------------------------
    # here's a brief explanation of how this works.
    # we're creating an array of 256 translation tables. This will become of
    # significance couple of lines later. In detail, this looks like this:
    # [
    #   ''.join([ 0x00 ^ 0x00, 0x00 ^ 0x01, 0x00 ^ 0x02, ...],
    #   ''.join([ 0x01 ^ 0x00, 0x01 ^ 0x01, 0x01 ^ 0x02, ...],
    #   ''.join([ 0x02 ^ 0x00, 0x02 ^ 0x01, 0x02 ^ 0x02, ...]
    #   ...
    #   ''.join([ 0xff ^ 0x00, 0xff ^ 0x01, 0xff ^ 0x02, ...]
    # ]
    # there are a total of 256 rows in this table, because there are a total of
    # 256 possible bytes.
    if six.PY2:
        _XOR_TABLE = [b''.join(chr(a ^ b) for a in range(256)) for b in range(256)]
    else:
        _XOR_TABLE = [bytes(a ^ b for a in range(256)) for b in range(256)]

    def mask(masking_key, data):
        """XOR mask bytes."""
        a, b, c, d = (_XOR_TABLE[n] for n in bytearray(masking_key))
        # there are 4 bytes in our masking key, that's why we are picking 4
        # variables from the table. Now, here comes the fun part. We are
        # converting `masking_key` to bytearray, which, when iterated over,
        # will covert a byte to the corresponding uint8_t, so if, for instance
        # the `masking_key` will be given as:
        # 'x07\x03\x01\x00'
        # then, n will have the value of 7, 3, 1, 0 with each pass of the
        # for-loop.
        # why is this so significant? Because, as you remember from above,
        # _XOR_TABLE has 256 rows, and the first byte in the xor operation
        # was changing from 0 to 256. Therefore, key from masking_key converted
        # to uint8_t can point us to a translation table for n-th byte from
        # `masking_key`.
        data_bytes = bytearray(data)
        data_bytes[::4] = data_bytes[::4].translate(a)
        data_bytes[1::4] = data_bytes[1::4].translate(b)
        data_bytes[2::4] = data_bytes[2::4].translate(c)
        data_bytes[3::4] = data_bytes[3::4].translate(d)
        # great. The rest is quite easy. array.translate expects a bytearray of
        # length=256. How convenient! It's exactly what we have. The way it
        # works is that it takes an input byte and looks for a byte in the
        # replacement table. So in our case, the replacement table will contain
        # XOR'ed value of this byte by the masking key. Now you may wonder -
        # why are there 4 iterations of this? Well, because there are 4
        # different translation tables for 4 bytes of our masking key - if we
        # wouldn't do this, then we would mess up our input data. So, for the
        # first byte of masking key, we do the following (O = leave original
        # byte, R = replace byte):
        # 0 1 2 3 4 5 6 7 8    ( byte no. )
        # R O O O R O O O R
        # then, for the second byte of the mask, we do:
        # 0 1 2 3 4 5 6 7 8    ( byte no. )
        # O R O O O R O O O
        # please note, that even though byte 0 is marked as 'O', it has already
        # been replaced in the previous step
        return bytes(data_bytes)
        # and voila!



"""
Manages individual Websocket frames.

A websocket 'message' may consist of several of these frames.

"""

from __future__ import print_function
from __future__ import unicode_literals

import struct

from . import errors
from .mask import make_masking_key, mask
from .opcode import is_reserved, Opcode


class Frame(object):
    """A raw websocket frame."""

    def __init__(self, opcode, payload=b'', masking_key=None,
                 fin=1, rsv1=0, rsv2=0, rsv3=0):
        self.opcode = opcode
        self.payload = payload
        self.masking_key = masking_key
        self.fin = fin
        self.rsv1 = rsv1
        self.rsv2 = rsv2
        self.rsv3 = rsv3

        self.mask = 0
        self.validate()

    def __repr__(self):
        opcode_name = Opcode.to_str(self.opcode)
        frame_type = "frame" if self.fin else "frame-fragment"
        return "<{} {} ({} bytes) fin={!r}>".format(
            frame_type,
            opcode_name,
            len(self),
            self.fin
        )

    def __len__(self):
        return len(self.payload)

    # Use struct module to pack ws frame header
    _pack8 = struct.Struct(b'!BB4s').pack  # 8 bit length field
    _pack16 = struct.Struct(b'!BBH4s').pack  # 16 bit length field
    _pack64 = struct.Struct(b'!BBQ4s').pack  # 64 bit length field
    _pack_close_code = struct.Struct(b'!H').pack

    @classmethod
    def build(cls, opcode, payload=b'',
              fin=1, rsv1=0, rsv2=0, rsv3=0,
              masking_key=None):
        """Build a WS frame header."""
        # https://tools.ietf.org/html/rfc6455#section-5.2
        masking_key = (
            make_masking_key()
            if masking_key is None
            else masking_key
        )
        mask_bit = 1 << 7
        byte0 = fin << 7 | rsv1 << 6 | rsv2 << 5 | rsv3 << 4 | opcode
        length = len(payload)
        if length < 126:
            header_bytes = cls._pack8(
                byte0, mask_bit | length, masking_key
            )
        elif length < (1 << 16):
            header_bytes = cls._pack16(
                byte0, mask_bit | 126, length, masking_key
            )
        elif length < (1 << 63):
            header_bytes = cls._pack64(
                byte0, mask_bit | 127, length, masking_key
            )
        else:  # pragma: no cover
            # Can't send a payload > 2**63 bytes
            raise errors.FrameBuildError(
                'payload is too large for a single frame'
            )
        frame_bytes = header_bytes + mask(masking_key, payload)
        return frame_bytes

    @classmethod
    def build_close_payload(cls, status, reason=''):
        """Build a close frame."""
        if not isinstance(reason, bytes):
            reason = reason.encode('utf-8', errors='replace')
        if status is None:
            return b''
        payload_bytes = cls._pack_close_code(status) + reason
        return payload_bytes

    def to_bytes(self):
        """Return binary encoding of WS frame."""
        frame_bytes = self.build(
            self.opcode,
            payload=self.payload
        )
        return frame_bytes

    def validate(self):
        """Check the frame and raise any errors."""
        if self.is_control and len(self.payload) > 125:
            raise errors.ProtocolError(
                "control frames must <= 125 bytes in length"
            )
        if self.rsv1 or self.rsv2 or self.rsv3:
            raise errors.ProtocolError(
                "reserved bits set"
            )
        if is_reserved(self.opcode):
            raise errors.ProtocolError(
                "opcode is reserved"
            )
        if not self.fin and self.is_control:
            raise errors.ProtocolError(
                "control frames may not be fragmented"
            )

    @property
    def is_control(self):
        """Check if this frame has a control opcode."""
        return self.opcode >= 8

    @property
    def is_text(self):
        """Check if this is a text frame."""
        return self.opcode == Opcode.TEXT

    @property
    def is_binary(self):
        """Check if this is a binary frame."""
        return self.opcode == Opcode.BINARY

    @property
    def is_continuation(self):
        """Check if this is a continuation."""
        return self.opcode == Opcode.CONTINUATION

    @property
    def is_ping(self):
        """Check if this is a ping frame."""
        return self.opcode == Opcode.PING

    @property
    def is_pong(self):
        """Check if this is a pong frame."""
        return self.opcode == Opcode.PONG

    @property
    def is_close(self):
        """Check if this is a close frame."""
        return self.opcode == Opcode.CLOSE


if __name__ == "__main__":  # pragma: no cover
    print(Frame(Opcode.BINARY, b'Hello, World', fin=0))
    print(Frame(Opcode.TEXT, b'Hello, World', fin=1))



from __future__ import print_function
from __future__ import unicode_literals

import logging
logging.basicConfig(format='%(message)s', level=logging.DEBUG)

from threading import Thread

from six.moves import input
from lomond import WebSocket

def get_input(text=''):
    i = input(text)
    if isinstance(i, bytes):
        return i.decode('utf-8', errors='replace')
    return i

name = get_input("your name: ")
ws = WebSocket('wss://ws.willmcgugan.com/chat/')

def run():
    for event in ws:
        print(event)
        if event.name == 'connected':
            ws.send_text("<{} connected>".format(name))
        elif event.name == 'text':
            print(event.text)

Thread(target=run).start()

while True:
    try:
        ws.send_text("[{}] {}".format(name, get_input()))
    except KeyboardInterrupt:
        ws.close()
        break



"""
Parses a websocket connection.

Yields a response object followed by 0 or more Websocket messages.

"""

from __future__ import unicode_literals

import logging

from six import text_type

from . import errors
from .frame_parser import ParseError, FrameParser
from .message import Message
from .response import Response


log = logging.getLogger('lomond')


class WebsocketStream(object):
    """
    Parses a stream of data in to Headers and logical Websocket
    frames.

    """

    def __init__(self):
        self.frame_parser = FrameParser()
        self._parsed_response = False
        self._frames = []

    def feed(self, data):
        """Feed in data from a socket to yield 0 or more frames."""
        # This combines fragmented frames in to a single frame
        iter_frames = iter(self.frame_parser.feed(data))

        if not self._parsed_response:
            # Process status line and headers from frame parser
            header_data = next(iter_frames, None)
            if header_data is None:
                return
            self._parsed_response = True
            yield Response(header_data)

        # Process incoming frames
        while True:
            try:
                frame = next(iter_frames)
            except ParseError as error:
                raise errors.CriticalProtocolError(
                    text_type(error)
                )
            log.debug("SRV -> CLI : %r", frame)
            if frame.is_control:
                # Control messages are never fragmented
                # And may be sent in the middle of a multi-part message
                yield Message.build([frame])
            else:
                # May be fragmented
                if frame.is_continuation and not self._frames:
                    raise errors.ProtocolError(
                        'continuation frame has nothing to continue'
                    )
                if not frame.is_continuation and self._frames:
                    raise errors.ProtocolError(
                        'continuation frame expected'
                    )
                self._frames.append(frame)
                if frame.fin:
                    yield Message.build(self._frames)
                    del self._frames[:]



"""

A message class, built from 1 or more websocket frames.

"""

from __future__ import unicode_literals

import struct

from . import errors
from .opcode import Opcode
from .utf8validator import Utf8Validator


class Message(object):
    """Base class for a websocket message.

    :param opcode: A message opcode defined in
        :class:`~lomond.opcode.Opcode`.

    """

    __slots__ = ['opcode']
    _unpack16 = struct.Struct(b'!H').unpack

    def __init__(self, opcode):
        self.opcode = opcode

    def __repr__(self):
        return "<message {}>".format(Opcode.to_str(self.opcode))

    @classmethod
    def build(cls, frames):
        """Build a message from a sequence of frames."""
        opcode = frames[0].opcode
        payload = b''.join(frame.payload for frame in frames)
        if opcode == Opcode.CLOSE:
            return Close.from_payload(payload)
        elif opcode == Opcode.PING:
            return Ping(payload)
        elif opcode == Opcode.PONG:
            return Pong(payload)
        elif opcode == Opcode.BINARY:
            return Binary(payload)
        elif opcode == Opcode.TEXT:
            return Text.from_payload(payload)
        else:
            return Message(opcode)

    @property
    def is_text(self):
        """Check if the message is text."""
        return self.opcode == Opcode.TEXT

    @property
    def is_binary(self):
        """Check if the message is binary."""
        return self.opcode == Opcode.BINARY

    @property
    def is_close(self):
        """Check if this is a close message."""
        return self.opcode == Opcode.CLOSE

    @property
    def is_ping(self):
        """Check if this is a ping message."""
        return self.opcode == Opcode.PING

    @property
    def is_pong(self):
        """Check if this is a pong message."""
        return self.opcode == Opcode.PONG


class Binary(Message):
    """Binary application data.

    :param bytes data: The message data.

    """
    __slots__ = ['data']
    def __init__(self, data):
        self.data = data
        super(Binary, self).__init__(Opcode.BINARY)

    def __repr__(self):
        return "<message BINARY {!r}>".format(self.data)


class Text(Message):
    """Text application data.

    :param str text: The message text.

    """
    __slots__ = ['text']
    def __init__(self, text):
        self.text = text
        super(Text, self).__init__(Opcode.TEXT)

    @classmethod
    def from_payload(cls, payload):
        try:
            text = payload.decode('utf-8')
        except UnicodeDecodeError as error:
            raise errors.CriticalProtocolError(
                'payload contains invalid utf-8; {}',
                error
            )
        return cls(text)

    def __repr__(self):
        return "<message TEXT {!r}>".format(self.text)


class Close(Message):
    """Connection close control message.

    :param int code: Close code.
    :param str reason: Close reason.

    """
    __slots__ = ['code', 'reason']
    def __init__(self, code, reason):
        self.code = code
        self.reason = reason
        super(Close, self).__init__(Opcode.CLOSE)

    @classmethod
    def from_payload(cls, payload):
        """Decode the error 'code' and 'reason'."""
        code = None
        reason = ''
        if len(payload) == 1:
            raise errors.ProtocolError(
                'invalid close frame payload'
            )
        elif len(payload) >= 2:
            (code,) = cls._unpack16(payload[:2])
            reason_bytes = payload[2:]
            is_valid, _, _, _ = Utf8Validator().validate(reason_bytes)
            if not is_valid:
                raise errors.CriticalProtocolError(
                    'close frame contains invalid utf-8'
                )
            try:
                reason = reason_bytes.decode('utf-8')
            except UnicodeDecodeError as error:
                raise errors.CriticalProtocolError(
                    'invalid utf-8 in close reason ({})',
                    error
                )
        return cls(code, reason)

    def __repr__(self):
        return "<message CLOSE {}, {!r}>".format(self.code, self.reason)


class Ping(Message):
    """Ping message.

    :param bytes data: Ping data.

    """
    __slots__ = ['data']
    def __init__(self, data):
        self.data = data
        super(Ping, self).__init__(Opcode.PING)

    def __repr__(self):
        return "<message PING {!r}>".format(self.data)


class Pong(Message):
    """Pong message.

    :param bytes data: Pong data.

    """
    __slots__ = ['data']
    def __init__(self, data):
        self.data = data
        super(Pong, self).__init__(Opcode.PONG)

    def __repr__(self):
        return "<message PONG {!r}>".format(self.data)
